/*
 * Copyright 2005, 2006 University of Leiden.
 *
 * This file is part of MIA+EWS.
 *
 * MIA+EWS is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MIA+EWS is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MIA; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#include "interpUtil.h"
/**********************************************************
 * Routines related to interpolation of an imaging_fdata_row
 * from the distorted to undistorted form.
 * Regridding is by convolution and then resampling onto 
 * the new grid,  this effectively convolved the output 
 * region images with the convolving function.
 ************************************************************* */

/* Global parameters */
#define CWIDTH 5               /* Width of convolving kernal in pixels */
#define GWIDTH 2.0             /* FWHM of Gaussian used for convolution */
#define NPERCELL 100           /* number of tabulated sample per cell */
#define CONVWD (CWIDTH+1)*NPERCELL  /* Size of convolution function table */
#define BLANK  0.0             /* Value to use for undefined value */
#define MINWEIGHT 0.35         /* Minimum summed weight in convolution */
#define MAXREGION 12           /* Maximum number of regions */

/* Global variables */
static float convfn[CONVWD+10];  /* tabulated convolving function, 
				    used in both dimensions */
static int convfnInit=0;         /* is the convolving function initialized?*/
static int numCoefX[MAXREGION];  /* number of nonzero coefficients in coefX */
static int numCoefY[MAXREGION];  /* number of nonzero coefficients in coefY */
static float coefX[MAXREGION][MAXCOEF];     /* coefficients of X distortion as
				    const, x, y, x^x, xy, y*y... */
static float coefY[MAXREGION][MAXCOEF];     /* coefficients of Y distortion as
				    const, x, y, x*x, xy, y*y... */

/* Local prototypes */
static void InitConvfn(void);    /* init convolving function */
/* initialize dedistortion polynomials */
static void InitCoef(imaging_detector_table_def *detect);
/* get dedistorted pixel location */
static float dedistortPixel(int xy, int ireg, float x, float y);
 
/***************************************************************
 * Dedistort the regions in rowin and put in rowout
 * Regridding is by convolution and then resampling onto 
 * the new grid,  this effectively convolved the output 
 * region images with the convolving function.
 * Convolution is flux conserving.
 * Inputs:
 *    detect IMAGING_DETECTOR structure describing input and output
 *    region region (0-rel) to dedistort, -1=> all.
 *    rowin  Data in distorted form
 *    rowout Dedistorted version
 **************************************************************** */
void minrtsDedistortFrame (imaging_detector_table_def *detect, int region,
			   imaging_fdata_row *rowin, imaging_fdata_row *rowout)
{
    int nregion, ireg, reglo, reghi, nx, ix, ny, iy, ipix, npix, xcell, ycell;
    int nperCell, delta, center, xcen, ycen, offset, jx, jy, kx, kkx, ky;
    int indx, ondx;
    float *weight=NULL;        /* work array for sum of weights */
    float *sumdata=NULL;       /* work array for sum of weights*data */
    float *idata, *odata, x, y, xx, yy, edge, wt; 

    /* Initialize convolvinfg function if needed */
    if (!convfnInit) InitConvfn();

    /* Set coefficients in dedistortion arrays */
    InitCoef(detect);

    /* loop over selected regions */
    nregion = detect->nregion;
    if (region>=0) {
      reglo = region;
      reghi = region+1;
    } else {
      reglo = 0;
      reghi = nregion;
    }
    for (ireg=reglo; ireg<reghi; ireg++) {

      /* Work space */
      nx = detect->naxis[ireg][0];
      ny = detect->naxis[ireg][1];
      npix = nx * ny;
      weight  = (float*)malloc(npix*sizeof(float));
      sumdata = (float*)malloc(npix*sizeof(float));

      /* init */
      idata = rowin->data[ireg];
      odata = rowout->data[ireg];
      for (ipix=0; ipix<npix; ipix++) {
	weight[ipix] = 0.0;
	sumdata[ipix] = 0.0;
      }

      /* Loop over input region */
      edge = (float)(CWIDTH / 2);
      nperCell = NPERCELL;
      offset = nperCell * (CWIDTH / 2);
      center = (1 + CWIDTH/2) * nperCell;
      for (iy=0; iy<ny; iy++) {
	yy = (float)(iy);
	for (ix=0; ix<nx; ix++) {
	  xx = (float)(ix);
	  /* Get dedistorted pixel location */
	  x = dedistortPixel (1, ireg, xx, yy);
	  y = dedistortPixel (2, ireg, xx, yy);

	  /* Avoid edge */
	  if ((y<edge) || (y>(ny-edge-1.0))) continue;

	  /* Get center tabulated value cell */
	  xcell = (int)(x+0.5);
	  if ((xcell<edge) || (xcell>(nx-edge-1.0))) continue;  /* Avoid edge */
	  /* get offset in table cells from being centered */
	  delta = -((float)xcell - x) * nperCell + 0.5;
	  xcen = center - delta - offset;
	  kkx = xcell - (CWIDTH / 2);

	  ycell = (int)(y+0.5);
	  if ((ycell<edge) || (ycell>(ny-edge-1.0))) continue;  /* Avoid edge */
	  delta = -((float)ycell - y) * nperCell + 0.5;
	  ycen = center - delta - offset;
	  ky = ycell - (CWIDTH / 2);

	  /* Loop over x, y regampling */
	  indx = (iy*nx) + ix;
	  for (jy = 0; jy<CWIDTH; jy++) {
	    kx = kkx;
	    for (jx = 0; jx<CWIDTH; jx++) {
	      ondx = (ky*nx) + kx;
	      wt = convfn[jy*nperCell+ycen] * convfn[jx*nperCell+xcen];
	      weight[ondx]  += wt;
	      sumdata[ondx] += idata[indx] * wt;
	      kx++;   /* Increment output x cell */
	    } /* end x resampling loop */
	    ky++;     /* Increment output y cell */
	  } /* end y resampling loop */

	} /* end x gridding loop */
      }   /* end y gridding loop */

      /* normalize */
      for (ipix=0; ipix<npix; ipix++) {
	if (weight[ipix]>MINWEIGHT) {
	  odata[ipix] = sumdata[ipix] / weight[ipix];
	} else odata[ipix] = BLANK;
      }

      /* cleanup */
      if (weight) free (weight);
      if (sumdata) free (sumdata);
    } /* end loop over regions */
} /* end  minrtsDedistortFrame */


/***************************************************************
 * Set dedistortion polynomials in IMAGING_DETECTOR structure 
 * Inputs:
 *    detect   IMAGING_DETECTOR structure describing input,
 *    iXY      0=>X, 1=>Y polynomial
 *    iReg     Region number
 *    ncoef    Number of coefficients
 *    coef     polynomial coefficients as
 *             const, x, y, x*x, x*x*x, x*y...
 **************************************************************** */
void minrtsSetPoly (imaging_detector_table_def *detect, int iXY, int iReg,
	      int ncoef, float *coef)
{
  int j;

    /* Initialize imaging_detector distortion terms */
    for (j=0; j<detect->max_coef; j++) {
      detect->dmp[iReg][j][iXY][0] = 0;
      detect->dmp[iReg][j][iXY][1] = 0;
      detect->dmc[iReg][j][iXY]   = -1.0e9; 
    }

    /* Put non zero entries in def */
    if (ncoef>=1) {
      detect->dmp[iReg][0][iXY][0] = 0;
      detect->dmp[iReg][0][iXY][1] = 0;
      detect->dmc[iReg][0][iXY]    = coef[0]; 
    }
    if (ncoef>=2) {
      detect->dmp[iReg][1][iXY][0] = 1;
      detect->dmp[iReg][1][iXY][1] = 0;
      detect->dmc[iReg][1][iXY]    = coef[1]; 
    }
    if (ncoef>=3) {
      detect->dmp[iReg][2][iXY][0] = 0;
      detect->dmp[iReg][2][iXY][1] = 1;
      detect->dmc[iReg][2][iXY]    = coef[2]; 
    }
    if (ncoef>=4) {
      detect->dmp[iReg][3][iXY][0] = 2;
      detect->dmp[iReg][3][iXY][1] = 0;
      detect->dmc[iReg][3][iXY]    = coef[3]; 
    }
    if (ncoef>=5) {
      detect->dmp[iReg][4][iXY][0] = 3;
      detect->dmp[iReg][4][iXY][1] = 0;
      detect->dmc[iReg][4][iXY]    = coef[4]; 
    }
    if (ncoef>=6) {
      detect->dmp[iReg][5][iXY][0] = 1;
      detect->dmp[iReg][5][iXY][1] = 1;
      detect->dmc[iReg][5][iXY]    = coef[5]; 
    }

} /* end minrtsSetPoly */


/***************************************************************
 * Get dedistortion polynomials from IMAGING_DETECTOR structure 
 * Inputs:
 *    detect   IMAGING_DETECTOR structure describing input,
 *    iXY      0=>X, 1=>Y polynomial
 *    iReg     Region number
 * Outputs:
 *    ncoef    Number of coefficients
 *    coef     polynomial coefficients as
 *             const, x, y, x*x, x*x*x, x*y...
 **************************************************************** */
void minrtsGetPoly (imaging_detector_table_def *detect, int iXY, int iReg,
	      int *ncoef, float *coef)
{
  int icoef, numCoef;

  /* Loop over regions */
  /* look for largest nonzero coef */
  numCoef = 0;
  for (icoef=0; icoef<detect->max_coef; icoef++) {
    /* look for specific terms */
    if (detect->dmc[iReg][icoef][iXY]>-1.0e8) { /* valid numbers */
      /* constant */
      if ((detect->dmp[iReg][icoef][iXY][0]==0) && (detect->dmp[iReg][icoef][iXY][1]==0)){
	coef[0] = detect->dmc[iReg][icoef][iXY];
	if ((numCoef<1) && (fabs(coef[0])>0.0)) numCoef = 1;
	
	/* x */
      } else if ((detect->dmp[iReg][icoef][iXY][0]==1) && (detect->dmp[iReg][icoef][iXY][1]==0)){
	coef[1] = detect->dmc[iReg][icoef][iXY];
	if ((numCoef<2) && (fabs(coef[1])>0.0)) numCoef = 2;
	
	/* y */
      } else if ((detect->dmp[iReg][icoef][iXY][0]==0) && (detect->dmp[iReg][icoef][iXY][1]==1)){
	coef[2] = detect->dmc[iReg][icoef][iXY];
	if ((numCoef<3) && (fabs(coef[2])>0.0)) numCoef = 3;
	
	/* x*x */
      } else if ((detect->dmp[iReg][icoef][iXY][0]==2) && (detect->dmp[iReg][icoef][iXY][1]==0)){
	coef[3] = detect->dmc[iReg][icoef][iXY];
	if ((numCoef<4) && (fabs(coef[3])>0.0)) numCoef= 4;
	
	/* x*x*x */
      } else if ((detect->dmp[iReg][icoef][iXY][0]==3) && (detect->dmp[iReg][icoef][iXY][1]==0)){
	coef[4] = detect->dmc[iReg][icoef][iXY];
	if ((numCoef<5) && (fabs(coef[4])>0.0)) numCoef= 5;
	
	/* x*y */
      } else if ((detect->dmp[iReg][icoef][iXY][0]==1) && (detect->dmp[iReg][icoef][iXY][1]==1)){
	coef[5] = detect->dmc[iReg][icoef][iXY];
	if ((numCoef<6) && (fabs(coef[5])>0.0))numCoef = 6;
      }
    }
    
  } /* end loop over coefficients */
  
  *ncoef = numCoef;
} /* end minrtsGetPoly */

/***************************************************************
 * Private: Tabulate convolution function
 * Modifies file globals convfnInit and convfn.
 **************************************************************** */
static void InitConvfn(void)
{
  int i, j, nperCell, nhalf;
  float arg, fact, sigma;

  convfnInit = 1;  /* now initialized */

  /* Tabulate colvolving function every (CONVWD/CWIDTH)-1 cells */
  /* Do lower half and later copy */
  nperCell = NPERCELL;
  nhalf = 1 + (1 + CWIDTH/2) * nperCell;
  /* Use Gaussian function */
  sigma = GWIDTH / 2.355;
  fact = - 1.0/ (sigma * sigma *((float)nperCell)*((float)nperCell));
  for (i=0; i<nhalf; i++) {
    arg = fact * (float)(nhalf - i - 1) * (float)(nhalf - i - 1);
    if (arg>-15.0) convfn[i] = exp (arg);
    else convfn[i] = 0.0;
  } /* end loop tabulating lower half */

  /* copy reflection of lower half to upper */
  j = nhalf - 2;
  for (i=nhalf; i<(nhalf*2 - 2); i++) {
    convfn[i] = convfn[j--];
  } /* end loop copying */
  
} /* end InitConvfn */


/***************************************************************
 * Initialize dedistortion coefficients
 * Modifies file globals, numCoefX, numCoefY, coefX, coefY
 * Inputs:
 **************************************************************** */
static void InitCoef(imaging_detector_table_def *detect)
{
  int ireg;

  /* Loop over regions */
  for (ireg=0; ireg<detect->nregion; ireg++) {
    minrtsGetPoly (detect, 0, ireg, &numCoefX[ireg], &coefX[ireg][0]);
    minrtsGetPoly (detect, 1, ireg, &numCoefY[ireg], &coefY[ireg][0]);
  } /* end loop over regions */

} /* end InitCoef */

/***************************************************************
 * Determine undistorted pixel location
 * Uses file globals, numCoef, numCoefY, coefX, coefY
 * Inputs:
 *    xy     1= return "X" pixel, 2=>"Y" pixel
 *    ireg   0-rel region number
 *    x      "X" in distorted image
 *    y      "Y" in distorted image
 * Returns:
 *    x or y dedistorted pixel.
 **************************************************************** */
/* get dedistorted pixel location */
static float dedistortPixel(int xy, int ireg, float x, float y)
{

  float sum, mpt[MAXCOEF];
  int i;

  /* Set polynomial terms in mpt  */
  SETMPT(mpt,x,y);

  /* expansion */
  sum = 0.0;
  if (xy==1) {  /* X */
   sum = x;
   for (i=0; i<numCoefX[ireg]; i++) 
      sum += mpt[i]*coefX[ireg][i];
  } else {      /* Y */
   sum = y;
   for (i=0; i<numCoefY[ireg]; i++) sum += mpt[i]*coefY[ireg][i];
  }
    return sum;   
} /* end dedistortPixel */

